!=========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: rangeindexreadstr.gs,v 1.7 2008-01-09 19:02:35 stever Exp $
!
! Superclass Hierarchy:
!   RangeIndexReadStream, BtreeReadStream, Stream, Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods RangeIndexReadStream
removeallclassmethods RangeIndexReadStream
protectmethods

category: 'For Documentation Installation only'
classmethod: RangeIndexReadStream
installDocumentation

< unprotected >

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'RangeIndexReadStream, like its superclass BtreeReadStream, supports the
 composition of query results by providing access to a B-tree structure.  Its
 ''next'' and ''atEnd'' methods are used the same way as those of
 BtreeReadStream in iterating through the B-tree.

 RangeIndexReadStream differs from BtreeReadStream in that it uses the reverse
 mappings to B-tree nodes that are found in a RangeEqualityIndex to obtain the
 next entry.  You can supply that index when you create the stream, and the
 index identifies the ordering used to return the entries.'.
doc documentClassWith: txt.

txt := (GsDocText new) details:
'The RangeEqualityIndex for an instance of this class.'.
doc documentInstVar: #rangeIndex with: txt.

txt := (GsDocText new) details:
'An Array of Integers that indicates the offset into BucketValueBags
 along the path.'.
doc documentInstVar: #setIterationIndexes with: txt.

self description: doc.
%

! ------------------- Class methods for RangeIndexReadStream
category: 'Instance Creation'
classmethod: RangeIndexReadStream
new

"Create an initialized instance of the receiver."
<protected>

| newOne |
newOne := super new.
newOne setIterationIndexes: Array new.
^ newOne
%

category: 'Instance Creation'
classmethod: RangeIndexReadStream
on: aRangeIndex

"Create a stream that can access the entire contents of the given
 RangeEqualityIndex."
<protected>

| newStream |
newStream := super on: aRangeIndex btreeRoot.
newStream rangeIndex: aRangeIndex.
^ newStream
%

! ------------------- Instance methods for RangeIndexReadStream
category: 'Private'
method: RangeIndexReadStream
_advance

"Advances the B-tree portion of the stream without advancing any of the set
 iteration indexes."
<protected>

^ super next
%

category: 'Testing'
method: RangeIndexReadStream
_atEnd

"Returns true if there are no more elements to return through the logical
 iteration of the stream."
<protected>

^ super atEnd _and: [ self noMoreSetsToIterate ]
%

category: 'Testing'
method: RangeIndexReadStream
atEnd

"Returns true if there are no more elements to return through the logical
 iteration of the stream."

<unprotected primitive: 901>
| atEnd |
atEnd := self _atEnd.
System _disableProtectedMode.
^ atEnd
%

category: 'Private'
method: RangeIndexReadStream
getNext

"Returns the next value on a stream of range index values."
<protected>

| val lastTerm |

" get the initial value at the end of the path
(but do not advance the stream yet) "
val := self _peekValue.

" if the index is not on the elements of the root NSC "
rangeIndex isIndexOnRootNsc
  ifFalse: [
    lastTerm := rangeIndex at: rangeIndex size.

    " if the last path term indicates that the index is on the elements of
    the NSC itself, then get the next value from the last term (the NSC)"
    lastTerm indicatesNsc
      ifTrue: [ val := self nextValueFor: val term: lastTerm ].

    " now iterate through the path components to get the final value "
    rangeIndex lastPathComponentsDictionaryOffset downTo: 1 do: [ :i |
      val := self nextValueFor: val term: (rangeIndex at: i)
    ].
  ].
" advance the B-tree current position "
super next.
^ val
%

category: 'Accessing'
method: RangeIndexReadStream
next

"Returns the next value on a stream of range index values."

<unprotected primitive: 901>
| result |
self _atEnd
    ifTrue: [ self _errorEndOfStream ].

result := self getNext.
System _disableProtectedMode.
^ result
%

category: 'Private'
method: RangeIndexReadStream
nextValueFor: val term: pathTerm

"Returns the next value on a stream of range index values."
<protected>

| nextVal |

nextVal := rangeIndex indexDictionary at: val term: pathTerm.
" see if the next value is a BucketValueBag "
(BucketValueBag _hasInstance: nextVal)
  ifTrue: [ nextVal := self nextValueInBag: nextVal level: pathTerm offset ].

^ nextVal
%

category: 'Accessing'
method: RangeIndexReadStream
rangeIndex

"Returns the value of the instance variable 'rangeIndex'."
<protected>

^rangeIndex
%

category: 'Updating'
method: RangeIndexReadStream
rangeIndex: newValue

"Modify the value of the instance variable 'rangeIndex'."
<protected>

rangeIndex := newValue.
self initializeSetIterationIndexes
%

category: 'Accessing'
method: RangeIndexReadStream
setIterationIndexes

"Returns the value of the instance variable 'setIterationIndexes'."
<protected>

^setIterationIndexes
%

category: 'Updating'
method: RangeIndexReadStream
setIterationIndexes: newValue

"Modify the value of the instance variable 'setIterationIndexes'."
<protected>

setIterationIndexes := newValue
%

category: 'Private'
method: RangeIndexReadStream
nextValueInBag: aBag level: offset

"Get the next value in the given BucketValueBag and update the set iteration
 indexes accordingly."
<protected>

| setOffset |

setOffset := (setIterationIndexes at: offset) + 1.

setIterationIndexes at: offset put:
  (setOffset == aBag size ifTrue: [ 0 ] ifFalse: [ setOffset ]).

" return the object in the set "
^ aBag _at: setOffset.
%

category: 'Query Support'
method: RangeIndexReadStream
makeNsc

"Returns a new NSC that contains all the elements from the receiver."

<unprotected primitive: 901>
| tmpHolder |
tmpHolder := NscBuilder for: rangeIndex nscRoot speciesForSelect new max: rangeIndex nscRoot size.

" If B-tree contains mapping to root object "
rangeIndex size == 1
    ifTrue: [ self _valuesAddInto: tmpHolder ]
    ifFalse: [ self _uniqueValuesAddInto: tmpHolder ].

" make the stream be at the end "
currentStack at: 1 put: 0.

System _disableProtectedMode.
^ tmpHolder completeBag
%

category: 'Query Support'
method: RangeIndexReadStream
_uniqueValuesAddInto: collection

"Iterates over the unique values of the B-tree, traversing the reverse mappings
 to the root object.  Adds the root to the given NSC."
<protected>

| leaf index eSize val prevVal offset pathTerm |

" get the index into the leaf node and see if it has reached the end "
index := currentStack at: 1.
index == 0
    ifTrue: [ ^ self ].

rangeIndex isIndexOnNscElements
    ifTrue: [ offset := rangeIndex size ]
    ifFalse: [ offset := rangeIndex lastPathComponentsDictionaryOffset max: 1 ].
pathTerm := rangeIndex at: offset.

leaf := currentStack at: 2.
eSize := endNode entrySize.
prevVal := #_incompletePathTraversal.
" while current node is not the end node "
[ leaf ~~ endNode ] whileTrue: [
    " iterate through the values of the current node "
    index to: leaf numElements * eSize by: eSize do: [ :i |
        val := leaf _basicAt: i.
        (val == prevVal)
            ifFalse: [
                rangeIndex _addAllFor: val
                into: collection
                offset: offset
                pathTerm: pathTerm
            ].
        prevVal := val
    ].

    " look down the stack for the next leaf node "
    self _nextLeaf.

    leaf := currentStack at: 2.
    index := 1
].
" iterate through the values of the end node "
index to: endIndex by: eSize do: [ :i |
    val := leaf _basicAt: i.
    (val == prevVal)
        ifFalse: [
            rangeIndex _addAllFor: val
            into: collection
            offset: offset
            pathTerm: pathTerm
        ].
    prevVal := val
].
%

category: 'Query Support'
method: RangeIndexReadStream
_uniqueValuesAddInto: collection identicalTo: aKey

"Iterate over the unique values of the B-tree, traversing the reverse mappings
 to the root object.  For each entry whose key is identical to the given key,
 add the root to the given NSC."
<protected>

| leaf index eSize key val prevVal offset pathTerm |

" get the index into the leaf node and see if it has reached the end "
index := currentStack at: 1.
index == 0
    ifTrue: [ ^ self ].

rangeIndex isIndexOnNscElements
    ifTrue: [ offset := rangeIndex size ]
    ifFalse: [ offset := rangeIndex lastPathComponentsDictionaryOffset max: 1 ].
pathTerm := rangeIndex at: offset.

leaf := currentStack at: 2.
eSize := endNode entrySize.
prevVal := #_incompletePathTraversal.
" while current node is not the end node "
[ leaf ~~ endNode ] whileTrue: [
    " iterate through the values of the current node "
    index to: leaf numElements * eSize by: eSize do: [ :i |
        key := leaf _basicAt: i + 1.
        val := leaf _basicAt: i.

        ( (key == aKey) _and:
        [ (val ~~ prevVal) ] )
            ifTrue: [
                rangeIndex _addAllFor: val
                into: collection
                offset: offset
                pathTerm: pathTerm
            ].
        prevVal := val
    ].

    " look down the stack for the next leaf node "
    self _nextLeaf.

    leaf := currentStack at: 2.
    index := 1
].
" iterate through the values of the end node "
index to: endIndex by: eSize do: [ :i |
    key := leaf _basicAt: i + 1.
    val := leaf _basicAt: i.

    ( (key == aKey) _and:
    [ (val ~~ prevVal) ] )
        ifTrue: [
            rangeIndex _addAllFor: val
            into: collection
            offset: offset
            pathTerm: pathTerm
        ].
    prevVal := val
].
%

category: 'Query Support'
method: RangeIndexReadStream
_uniqueValuesAddInto: collection notIdenticalTo: aKey

"Iterate over the unique values of the B-tree, traversing the reverse mappings
 to the root object.  For each entry whose key is not identical to the given
 key, add the root to the given NSC."
<protected>

| leaf index eSize key val prevVal offset pathTerm |

" get the index into the leaf node and see if it has reached the end "
index := currentStack at: 1.
index == 0
    ifTrue: [ ^ self ].

rangeIndex isIndexOnNscElements
    ifTrue: [ offset := rangeIndex size ]
    ifFalse: [ offset := rangeIndex lastPathComponentsDictionaryOffset max: 1 ].
pathTerm := rangeIndex at: offset.

leaf := currentStack at: 2.
eSize := endNode entrySize.
prevVal := #_incompletePathTraversal.
" while current node is not the end node "
[ leaf ~~ endNode ] whileTrue: [
    " iterate through the values of the current node "
    index to: leaf numElements * eSize by: eSize do: [ :i |
        key := leaf _basicAt: i + 1.
        val := leaf _basicAt: i.
        ( (key ~~ aKey) _and:
        [ (val ~~ prevVal) ] )
            ifTrue: [
                rangeIndex _addAllFor: val
                into: collection
                offset: offset
                pathTerm: pathTerm
            ].
        prevVal := val
    ].

    " look down the stack for the next leaf node "
    self _nextLeaf.

    leaf := currentStack at: 2.
    index := 1
].
" iterate through the values of the end node "
index to: endIndex by: eSize do: [ :i |
    key := leaf _basicAt: i + 1.
    val := leaf _basicAt: i.
    ( (key ~~ aKey) _and:
    [ (val ~~ prevVal) ] )
        ifTrue: [
            rangeIndex _addAllFor: val
            into: collection
            offset: offset
            pathTerm: pathTerm
        ].
    prevVal := val
].
%

category: 'Query Support'
method: RangeIndexReadStream
approxNumElements

"Returns the number of leaf node entries represented by the receiver.  This is
 approximately the number of elements because objects may share sub-objects
 which converge to the same leaf node entry."

<unprotected primitive: 901>
| result |
self atEnd
  ifTrue: [
    System _disableProtectedMode.
    ^ 0
  ].

result := Array with: 0.

(currentStack at: currentStack size) numElementsInStack: currentStack 
    level:  currentStack size - 1
    endNode: endNode
    endIndex: endIndex 
    into: result.

System _disableProtectedMode.
" Add one because the last entry is not counted "
^ (result at: 1) + 1
%

category: 'Copying'
method: RangeIndexReadStream
copy

"Returns a copy of the receiver."

<unprotected>

| newOne |

newOne := super copy.
newOne currentStack: currentStack copy.
^ newOne
%

category: 'Query Support'
method: RangeIndexReadStream
makeNscFilterSymbols: bool

"Returns a new NSC that contains all the elements from the receiver.  If bool is
 true, only Symbols are to be placed in the result; if bool is false, no Symbols
 are to be placed in the result."

<unprotected primitive: 901>

| tmpHolder |

tmpHolder := NscBuilder for: rangeIndex nscRoot speciesForSelect new max: rangeIndex nscRoot size.

bool
  ifTrue: [
    " If B-tree contains mapping to root object "
    rangeIndex size == 1
      ifTrue: [ self _valuesOnlySymbolsAddInto: tmpHolder ]
      ifFalse: [ self _uniqueValuesOnlySymbolsAddInto: tmpHolder ].
  ]
  ifFalse: [
    " If B-tree contains mapping to root object "
    rangeIndex size == 1
      ifTrue: [ self _valuesNoSymbolsAddInto: tmpHolder ]
      ifFalse: [ self _uniqueValuesNoSymbolsAddInto: tmpHolder ].
  ].

" make the stream be at the end "
currentStack at: 1 put: 0.

System _disableProtectedMode.
^ tmpHolder completeBag
%

category: 'Query Support'
method: RangeIndexReadStream
_uniqueValuesOnlySymbolsAddInto: collection

"Iterates over the unique values that are Symbols, traversing the reverse
 mappings to the root object.  Adds the root to the given NSC."

<protected>

| leaf index eSize val prevVal offset pathTerm |

" get the index into the leaf node and see if it has reached the end "
index := currentStack at: 1.
index == 0
  ifTrue: [ ^ self ].

rangeIndex isIndexOnNscElements
  ifTrue: [ offset := rangeIndex size ]
  ifFalse: [ offset := rangeIndex lastPathComponentsDictionaryOffset max: 1 ].
pathTerm := rangeIndex at: offset.

leaf := currentStack at: 2.
eSize := endNode entrySize.
prevVal := #_incompletePathTraversal.
" while current node is not the end node "
[ leaf ~~ endNode ] whileTrue: [
  " iterate through the values of the current node "
  index to: leaf numElements * eSize by: eSize do: [ :i |
    ((leaf _isSymbolAt: i+1) _and:
    [ (val := leaf _basicAt: i) ~~ prevVal ])
      ifTrue: [
        rangeIndex _addAllFor: val
          into: collection
          offset: offset
          pathTerm: pathTerm
      ].
    prevVal := val
  ].

  " look down the stack for the next leaf node "
  self _nextLeaf.

  leaf := currentStack at: 2.
  index := 1
].
" iterate through the values of the end node "
index to: endIndex by: eSize do: [ :i |
  ((leaf _isSymbolAt: i+1) _and:
  [ (val := leaf _basicAt: i) ~~ prevVal ])
    ifTrue: [
      rangeIndex _addAllFor: val
        into: collection
        offset: offset
        pathTerm: pathTerm
    ].
  prevVal := val
].
%

category: 'Query Support'
method: RangeIndexReadStream
_uniqueValuesNoSymbolsAddInto: collection

"Iterates over the unique values that are Symbols, traversing the reverse
 mappings to the root object.  Adds the root to the given NSC."

<protected>

| leaf index eSize val prevVal offset pathTerm |

" get the index into the leaf node and see if it has reached the end "
index := currentStack at: 1.
index == 0
  ifTrue: [ ^ self ].

rangeIndex isIndexOnNscElements
  ifTrue: [ offset := rangeIndex size ]
  ifFalse: [ offset := rangeIndex lastPathComponentsDictionaryOffset max: 1 ].
pathTerm := rangeIndex at: offset.

leaf := currentStack at: 2.
eSize := endNode entrySize.
prevVal := #_incompletePathTraversal.
" while current node is not the end node "
[ leaf ~~ endNode ] whileTrue: [
  " iterate through the values of the current node "
  index to: leaf numElements * eSize by: eSize do: [ :i |
    ((leaf _isSymbolAt: i+1) _or:
    [ (val := leaf _basicAt: i) == prevVal ])
      ifFalse: [
        rangeIndex _addAllFor: val
          into: collection
          offset: offset
          pathTerm: pathTerm
      ].
    prevVal := val
  ].

  " look down the stack for the next leaf node "
  self _nextLeaf.

  leaf := currentStack at: 2.
  index := 1
].
" iterate through the values of the end node "
index to: endIndex by: eSize do: [ :i |
  ((leaf _isSymbolAt: i+1) _or:
  [ (val := leaf _basicAt: i) == prevVal ])
    ifFalse: [
      rangeIndex _addAllFor: val
        into: collection
        offset: offset
        pathTerm: pathTerm
    ].
  prevVal := val
].
%

category: 'Private'
method: RangeIndexReadStream
initializeSetIterationIndexes

"Initialize the setIterationIndexes for a new instance."

<protected>
rangeIndex == nil
  ifTrue: [ ^ self ].

setIterationIndexes := Array new: rangeIndex lastPathComponentsDictionaryOffset.
1 to: setIterationIndexes size do: [ :i | setIterationIndexes at: i put: 0 ].
%

category: 'Private'
method: RangeIndexReadStream
noMoreSetsToIterate

"Return whether the setIterationIndexes specifies any more sets over
which to iterate."
<protected>

1 to: setIterationIndexes size do: [ :i |
  (setIterationIndexes at: i) > 0
    ifTrue: [ ^ false ]
].
^ true
%

unprotectmethods

